/* 
 *                     CopyRight (C) 1995 Harm Arts
 *                          Copying-policy: GPL
 *                    email: harm@chaos.es.ele.tue.nl
 *
 * xmemuse.c      -- show memory usage (a la xload)
 *
 * Created On      : Mon May  8 23:29:53 1995
 * Last Modified On: Wed Apr 17 10:43:39 1996
 *
 * $Id: xmemuse.c,v 0.4 1996/04/17 08:18:18 harm Exp harm $
 *
 *****************************************************************************/

#include <stdio.h>
#include <limits.h>
#include <stdarg.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Simple.h>

#define MEMINFO		"/proc/meminfo"
#define CLASSTYPE	"XMemuse"
#define NOMP		2048
#define MEMPOINTS	6
#define MEMPARTS	4

typedef struct {
  int		scale;		/* draw scale lines */
  Pixel		sc;		/* scale lines */       
  int		update;		/* update interval in seconds */
  Pixel		sw;		/* swap line */         
  Pixel		xs;		/* not available swap */
  Pixel		fs;		/* free swap */         
  Pixel		um;		/* used ram + swap */   
  Pixel		fm;		/* free ram */          
  Pixel		cm;		/* cached mem */        
  Pixel		bm;		/* buffers ram */       
  Pixel		sm;		/* shared mem */        
#ifdef DEBUG
  int		debug;		/* print debug-level info */
#endif
} AppData;

unsigned int		max_mem = 0;
unsigned short		mempoints[NOMP][MEMPOINTS];
unsigned short		swapline;
int			base = 0;

double			shrt2hght;

Display 		*dpy;
GC			gc;
Window 			wnd;
XWindowAttributes	w_att;
int			expect_gexpo = 0;
AppData			app_data;

#define RESOURCE(N,T,W,V) { #N, XtC ## T, XtR ## T, sizeof(T), \
			      XtOffsetOf(AppData, N), XtR ## W, (XtPointer) V }

#define XtCInt		"Int"
#define XtCPixel	"Color"
#define Int		int

static XtResource resources[] = {
  RESOURCE (scale    , Int    , Immediate, 4       ),
  RESOURCE (sc       , Pixel  , String   , "gray73"),
  RESOURCE (update   , Int    , Immediate, 10      ),
  RESOURCE (sw       , Pixel  , String   , "black" ),
  RESOURCE (xs       , Pixel  , String   , "black" ),
  RESOURCE (fs       , Pixel  , String   , "green" ),
  RESOURCE (um       , Pixel  , String   , "red"   ),
  RESOURCE (fm       , Pixel  , String   , "green" ),
  RESOURCE (cm       , Pixel  , String   , "orange"),
  RESOURCE (bm       , Pixel  , String   , "blue"  ), 
  RESOURCE (sm       , Pixel  , String   , "yellow")
#ifdef DEBUG
 ,RESOURCE (debug    , Int    , Immediate, 0       )
#endif
};  

static XrmOptionDescRec options[] = {
  {"-scale",	"*scale",	XrmoptionSepArg,	NULL},
  {"-sc",	"*sc",		XrmoptionSepArg,	NULL},
  {"-update",	"*update",	XrmoptionSepArg,	NULL},
  {"-sw",	"*sw",		XrmoptionSepArg,	NULL},
  {"-xs",	"*xs",		XrmoptionSepArg,	NULL},
  {"-fs",	"*fs",		XrmoptionSepArg,	NULL},
  {"-um",	"*um",		XrmoptionSepArg,	NULL},
  {"-fm",	"*fm",		XrmoptionSepArg,	NULL},
  {"-cm",	"*cm",		XrmoptionSepArg,	NULL},
  {"-bm",	"*bm",		XrmoptionSepArg,	NULL}, 
  {"-sm",	"*sm",		XrmoptionSepArg,	NULL},
#ifdef DEBUG
  {"-debug",	"*debug",	XrmoptionSepArg,	NULL},
#endif
};

static String fallback_resources[] = {
  "*geometry: 200x100",
  NULL
};

void
Usage (int argc, char * argv[])
{
  int i;
  
  printf ("Wrong option%s : %s\n", (argc > 2)?"s":" ", argv[1]);
  for (i = 2; i < argc; i++)
    printf ("                %s\n", argv[i]);
  printf ("\nUsage: %s [toolkitoptions] [options]\n", argv[0]);
  for (i = 0; i < XtNumber (options); i++)
    printf ("%8s %s %s\n", (i == 0)?"Options:":"", options[i].option,
	    (options[i].argKind == XrmoptionSepArg)?"<value>":"");
  printf ("(The 2-letter options take a color as argument,\n"
	  " scale in Mbytes, update in seconds)\n");
  exit (1);
}


int
QuitX (const char *fmt, ...)
{
  va_list   ap;
  va_start (ap, fmt);
  fprintf  (stderr, "Error: ");
  vfprintf (stderr, fmt, ap);
  fprintf  (stderr, "\n");
  exit (1);
}

void
DrawLine (int mempos, int xpos)
{
  int			i;
  unsigned short	points[MEMPOINTS];
  int                   pmempos = ((mempos)? mempos : NOMP) - 1;

  for (i = 0; i< MEMPOINTS; i++)
    points[i] = mempoints[mempos][i] * shrt2hght;
#ifdef DEBUG
  if (app_data.debug > 2)
    printf ("DrawLine %3d - 0, %d, %d, %d, %d, %d, %d\n", xpos, points[0],
	    points[1], points[2], points[3], points[4], w_att.height);
#endif
  if (points[0] > 0)
    {
      XSetForeground (dpy, gc, app_data.xs);
      XDrawLine (dpy, wnd, gc, xpos, 0, xpos, points[0]);
    }
  if (points[1] > points[0])
    {
      XSetForeground (dpy, gc, app_data.fs);
      XDrawLine (dpy, wnd, gc, xpos, points[0], xpos, points[1]);
    }
  if (points[2] > points[1])
    {
      XSetForeground (dpy, gc, app_data.um);
      XDrawLine (dpy, wnd, gc, xpos, points[1], xpos, points[2]);
    }
  if (points[3] > points[2])
    {  
      XSetForeground (dpy, gc, app_data.fm);
      XDrawLine (dpy, wnd, gc, xpos, points[2], xpos, points[3]); 
    }
  if (points[4] > points[3])
    {  
      XSetForeground (dpy, gc, app_data.cm);
      XDrawLine (dpy, wnd, gc, xpos, points[3], xpos, points[4]); 
    }
  if (w_att.height > points[4])
    {  
      XSetForeground (dpy, gc, app_data.bm);
      XDrawLine (dpy, wnd, gc, xpos, points[4], xpos, w_att.height);
    }

  if (app_data.scale)
    {
      int i, su = app_data.scale * shrt2hght;

      XSetForeground (dpy, gc, app_data.sc);
      for (i = w_att.height; i > 0; i -= su)
	XDrawPoint (dpy, wnd, gc, xpos, i);
    }
  if (swapline)
    {
      XSetForeground (dpy, gc, app_data.sw);
      XDrawPoint (dpy, wnd, gc, xpos, (int) (swapline * shrt2hght));
    }
  
  XSetForeground (dpy, gc, app_data.sm);
  i = mempoints[pmempos][5] * shrt2hght;
  if (abs (i - points[5]) > 1)
    XDrawLine (dpy, wnd, gc, xpos, points[5], xpos, i);
  else
    XDrawPoint (dpy, wnd, gc, xpos, points[5]);
}

void
ReDrawWindow (int x, int y, int w, int h)
{
  int i;
  
#ifdef DEBUG
  if (app_data.debug > 1)
    printf ("Redraw: base=%d, x=%d, w=%d, ww=%d, y=%d, h=%d, wh=%d\n",
	    base,  x, w, w_att.width, y, h, w_att.height);
  if (x < 0 || w < 0 || x + w > w_att.width ||
      y < 0 || h < 0 || y + h > w_att.height)
    QuitX ("Wrong expose dimensions");
#endif
  
  for (i = x; i < x + w; i++)
    DrawLine ((NOMP + base - w_att.width + i) % NOMP, i);
}

void
GetMemPoints (unsigned short points[])
{
  /* uint ==> max. 4Gbyte ram + swap; should be enough for a few years ... */
  unsigned int  cumm, tm, um, fm, sm, bm, cm, ts, us, fs;
  FILE          *mi = NULL;
  char          buf[80];
  static double	norm;

  if (!(mi = fopen(MEMINFO, "r"))) QuitX ("Can't open meminfo");
  
  fgets(buf, 79, mi);
  fgets(buf, 79, mi);
  switch (sscanf (buf, "%*s %d %d %d %d %d %d", &tm, &um, &fm, &sm, &bm, &cm))
    {
    case 5: cm = 0;		/* kernel < 1.3.50 */
    case 6: break;		/* kernel > 1.3.50 */
    default: QuitX ("/proc/meminfo has changed format?");
    }
  fgets(buf, 79, mi);
  switch (sscanf (buf, "%*s %d %d %d"      , &ts, &us, &fs))
    {
    case 3: break;
    default: QuitX ("/proc/meminfo has changed format?");
    }
  
  if (fclose (mi)) QuitX ("Can't close meminfo");
  
  if (max_mem == 0) /* initialize */
    {
#ifdef DEBUG
      if (app_data.debug > 0)
	printf ("%9s %9s %9s %9s %9s %9s | %9s %9s %9s\n"
		"%9s %9s %9s %9s %9s %9s | %9s %9s %9s\n",
		"Mem:", "", "", "", "", "", "Swap:", "", "", 
		"total:", "used:", "free:", "shared:", "buffers:", "cached:",
		"total:", "used:", "free:");
#endif
      max_mem = tm + ts;
      norm = (max_mem) / (double) USHRT_MAX;
      swapline = ts / norm;
      app_data.scale *= 1024 * 1024 / norm;
    }
  else if (max_mem < tm + ts)	/* swap increased */
    {				/* (tm didn't change; I hope ...) */
      int i,j;			/* reinitialize */
      for (i = 0; i < NOMP; i++)
	for(j = 0; j < MEMPOINTS; j++)
	  mempoints[i][j] = USHRT_MAX - ((USHRT_MAX - mempoints[i][j]) *
					 ((max_mem / (double) (tm + ts))));
      app_data.scale *= (max_mem / (double) (tm + ts));
      max_mem = tm + ts;
      norm = (max_mem) / (double) USHRT_MAX;
      swapline = ts / norm;
      ReDrawWindow (0, 0, w_att.width, w_att.height);
    }
#ifdef DEBUG
  if (app_data.debug > 0)
    printf ("%9d %9d %9d %9d %9d %9d | %9d %9d %9d\n",
	      tm, um, fm, sm, bm, cm,   ts, us, fs);
#endif

  /* mempoints are ushort ==> 64k resolution for 100% window.height */
  um = tm - fm - cm - bm + us;
  points[0] = (cumm = max_mem - tm - ts) / norm;
  points[1] = (cumm += fs) / norm;
  points[2] = (cumm += um) / norm;
  points[3] = (cumm += fm) / norm;
  points[4] = (cumm += cm) / norm;
  points[5] = (max_mem - sm) / norm;
  if (cumm + bm != max_mem) QuitX ("cumm + bm != max_mem\n");
  return;
}

void
ScrollMap (void)
{
#ifdef DEBUG
  if (app_data.debug > 0) printf ("ScrollMap\n");
#endif

  GetMemPoints (mempoints[base]);

  XCopyArea (dpy, wnd, wnd, gc, 1, 0, w_att.width - 1, w_att.height, 0, 0);
  DrawLine (base, w_att.width - 1);
  expect_gexpo++;
  
  if (++base == NOMP)
    base = 0;
}

void
Update (XtPointer app_ctxt,  XtIntervalId *id)
{
  XtAppAddTimeOut (app_ctxt, (unsigned long)app_data.update * 1000,
		   Update, app_ctxt);

  ScrollMap ();
}

void
HandleKeyPress (Widget s_wdg, XtPointer unsused, XEvent *ev, Boolean *c_t_d)
{
  KeySym		ks;
  char			string[10];
  XComposeStatus	cs;

  XLookupString ((XKeyEvent *) ev, string, 9, &ks, &cs);
#ifdef DEBUG
  if (app_data.debug > 0) printf ("KeyPress EVENT: %ld\n", ks);
#endif
  switch (ks)
    {
    case XK_Q:
      exit (0);
    case XK_space:
      ScrollMap ();
    }
}

void
HandleExpose (Widget s_wdg, XtPointer unsused, XEvent *ev, Boolean *c_t_d)
{
#ifdef DEBUG
  if (app_data.debug > 0)
    printf ("Expose EVENT: %d %d %d %d: expect = %d\n",
	    ev -> xexpose.x    , ev -> xexpose.y,
	    ev -> xexpose.width, ev -> xexpose.height, expect_gexpo);
#endif
  ReDrawWindow ((ev -> xexpose.x > expect_gexpo)?
		ev -> xexpose.x     - expect_gexpo : 0, ev -> xexpose.y,
		ev -> xexpose.width                   , ev -> xexpose.height);
}

void
HandleGExpose (Widget s_wdg, XtPointer unsused, XEvent *ev, Boolean *c_t_d)
{
  if (ev -> type == GraphicsExpose)
    {
#ifdef DEBUG
      if (app_data.debug > 0)
	printf ("GraphicsExpose EVENT: %d %d %d %d, "
		"serno = %ld, count = %d, gexpo = %d\n",
		ev -> xgraphicsexpose.x        , ev -> xgraphicsexpose.y,
		ev -> xgraphicsexpose.width    , ev -> xgraphicsexpose.height,
		ev -> xgraphicsexpose.serial   , ev -> xgraphicsexpose.count,
		expect_gexpo);
#endif
      ReDrawWindow (ev -> xgraphicsexpose.x - expect_gexpo + 1
		                               , ev -> xgraphicsexpose.y,
		    ev -> xgraphicsexpose.width, ev -> xgraphicsexpose.height);
      if (ev -> xgraphicsexpose.count == 0)
	expect_gexpo--;
    }
  else if (ev -> type == NoExpose)
    {
#ifdef DEBUG
      if (app_data.debug > 0) printf ("NoExpose EVENT\n");
#endif
      expect_gexpo--;
    }
  else
    {
#ifdef DEBUG
      if (app_data.debug > 0)
	printf ("GExpose EVENT %d: %d %d %d %d\n", ev -> type,
		ev -> xexpose.x, ev -> xexpose.y,
		ev -> xexpose.width, ev -> xexpose.height);
#endif
      ReDrawWindow (ev -> xexpose.x    , ev -> xexpose.y,
		    ev -> xexpose.width, ev -> xexpose.height);
    }
}

void
HandleStructure (Widget s_wdg, XtPointer unsused, XEvent *ev, Boolean *c_t_d)
{
  if (ev -> type == ConfigureNotify)
    {
#ifdef DEBUG
      if (app_data.debug > 0)
	printf ("ConfigureNotify EVENT: %dx%d+%d+%d\n",
		ev -> xconfigure.width, ev -> xconfigure.height,
		ev -> xconfigure.x, ev -> xconfigure.y);
#endif      
      if (w_att.width  != ev -> xconfigure.width ||
	  w_att.height != ev -> xconfigure.height)
	{
	  w_att.width  = ev -> xconfigure.width;
	  w_att.height = ev -> xconfigure.height;
	  
	  shrt2hght = w_att.height / (double) USHRT_MAX;	  
#ifdef DEBUG
	  if (app_data.debug > 0) printf("RedrawWindow\n");
#endif
	  ReDrawWindow (0, 0, w_att.width, w_att.height);
	}
    }
  else if (ev -> type == MapNotify)
    {		/* if xdm started us, we wont get an Expose event */
#ifdef DEBUG
      if (app_data.debug > 0) printf ("MapNotify EVENT\n");
#endif      
      ReDrawWindow (0, 0, w_att.width, w_att.height);
    }
#ifdef DEBUG
  else if (app_data.debug > 0) printf ("Structure EVENT %d\n", ev -> type);
#endif      
}

int
main (int argc, char *argv[])
{
  int		i, j;
  Widget	toplevel, s_wdg;
  XtAppContext	app_ctxt;
  XGCValues	gcv;

  toplevel = XtVaAppInitialize (&app_ctxt, CLASSTYPE,
				options, XtNumber (options),
				&argc, argv, fallback_resources, NULL);
  if (argc > 1) Usage (argc, argv);
  XtGetApplicationResources (toplevel, &app_data,
			     resources, XtNumber(resources), NULL, 0);
  
  s_wdg = XtVaCreateManagedWidget ("simple", simpleWidgetClass, toplevel, NULL);

  XtRealizeWidget (toplevel);

  dpy = XtDisplay (s_wdg);
  wnd = XtWindow (s_wdg);

  XtAddEventHandler (s_wdg, KeyPressMask,        False, HandleKeyPress , NULL);
  XtAddEventHandler (s_wdg, ExposureMask,        False, HandleExpose   , NULL);
  XtAddEventHandler (s_wdg, StructureNotifyMask, False, HandleStructure, NULL);
  XtAddEventHandler (s_wdg, 0,                   True , HandleGExpose  , NULL);

  XGetWindowAttributes (dpy, wnd, &w_att);
#ifdef DEBUG
  if (app_data.debug > 0) printf ("Geometry = %dx%d+%d+%d\n",
				  w_att.width, w_att.height, w_att.x, w_att.y);
#endif  
  gcv.line_width = 1;
  gcv.line_style = 0;
  gc = XtGetGC (s_wdg, (GCLineWidth | GCLineStyle), &gcv);

  shrt2hght = w_att.height / (double) USHRT_MAX;
  
/*   while (!XtIsRealized (s_wdg)) */
/*     sleep (1); */

  GetMemPoints (mempoints[base = 0]);

  for (i = 1; i < NOMP; i++)
    for (j = 0; j < MEMPOINTS; j++)
      mempoints[i][j] = mempoints[0][j];
#ifdef DEBUG
  if (app_data.debug > 0) printf ("SetTimer\n");
#endif
#if 1	/* if xdm started this, you wont get an Expose event !? */

#endif
  XtAppAddTimeOut (app_ctxt, (unsigned long)app_data.update * 1000,
		   Update, app_ctxt);
#ifdef DEBUG
  if (app_data.debug > 0) printf ("EventLoop\n");
#endif
  XtAppMainLoop (app_ctxt);

  exit (0);
}
